/*/
Impulse.h See below for description.
Copyright (C) 1995  Billy Stiltner

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License,

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
/*/

//The classes herein can be used together to provide a basis for a
//response mechanism.
//
//The impulse is a node based object.
//
//The Reactor is a list based object.
//
//The Coordinator is used as a wrapper in that it maintains two Reactors,
//an ACTIVE Reactor, and an INACTIVE Reactor so that responses can be
//loaded and unloaded hence DYNAMIC RESPONSE LOADING (DRL).
//
//The Coordinator can make reflexes and responses.
//
//The impulse is a dual natured class, it can be a reflex or a response.
//
//The reflex is created with a pointer to a variable uint *Var, a uint
//Condition, and a pMF pFunc. HOW IT WORKS: reflex's VAR is set to Var,
//CONDITION is set to Condition, and RESPOND is set to pFunc, if *VAR,
//wherever it may be, is equal to CONDITION, whatever it may be, RESPOND,
//whatever it does, is CALLED.
//
//The response is created with a uint Message and a pMF pFunc. HOW IT WORKS:
//response's CONDITION is set to Message and RESPOND is set to pFunc, *VAR
//must be provided at response time if *VAR = CONDITION then RESPOND is CALLED.
//
//The Reactor maintains reflexes and responses, it can only hold one
//type of impulse, but can hold as many as there is memory to put them.
//
//
//include files//////////////////////////////////////////////////////////////
//#include<objects.h>
//#include<nexus.h>
//#include"c:\program\svga\objects.h"
/////////////////////////////////////////////////////////////////////////////

//defines////////////////////////////////////////////////////////////////////
#define TORPID 0L
#define AWAKE 1L
/////////////////////////////////////////////////////////////////////////////

//variables//////////////////////////////////////////////////////////////////
unsigned long impulseconstructors;
unsigned long coordinatorconstructors;
unsigned long impulsedestructors;
unsigned long coordinatordestructors;
/////////////////////////////////////////////////////////////////////////////

//classes////////////////////////////////////////////////////////////////////
class impulse:  public NEXUS
{
 public://access modifier
//data members///////////////////////////////////////////////////////////////
 pMF RESPOND;                   //*function to call if message = CONDITION
 unsigned int *VAR;             //unsigned int pointer to a variable to check
 unsigned int CONDITION;        //value to compare condition with
//member functions///////////////////////////////////////////////////////////
 //impulse();                             //default constructor sets all to NULL
 impulse(impulse *i);                   //constructer using impulse *
 impulse(unsigned int Message,          //consructor to make a response
	 pMF pFunc);
 impulse(unsigned int *Var,             //consructor to make a reflex
	 unsigned int Condition,
	 pMF pFunc);
 void makeresponse(unsigned int Message,//function to make a response
			pMF pFunc);
 void makereflex(unsigned int *Var,     //function to make a reflex
		 unsigned int Condition,
		 pMF pFunc);
 void respond(NEXUS *pN);               //function to call when responding

 ~impulse();                   //destructor
};
/////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
class Reactor : public ROTI
{
public:
 //data members
 //member functions
 Reactor();
 //void DefineResponse(impulse *im);
  impulse *DefineReflex(unsigned int *Var,
			 unsigned int Condition,
			 pMF pFunc);
  impulse *DefineResponse(unsigned int Message,
			  pMF pFunc);
  virtual void Reflex(NEXUS *pN);
  virtual void Respond(MSG *Msg,NEXUS *pN);
  ~Reactor();
};
////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////
class Coordinator
{
public:
Reactor *ACTIVE;
Reactor *INACTIVE;
Coordinator();
impulse *MakeReflex(unsigned int *Var,
			 unsigned int Condition,
			 pMF pFunc);
impulse *MakeResponse(unsigned int Message,
			  pMF pFunc);
void Activate(impulse *i);
void Deactivate(impulse *i);
virtual void Reflex(NEXUS *pN);
virtual void Respond(MSG *Msg,NEXUS *pN);
~Coordinator();
};
////////////////////////////////////////////////////////////////////

//SUGGESTION:
//make a member of class x

// x::MakeResponseTable()
// {
// DEFINE_RESPONSE(x,MessageId,NameOfFunctionCalledInResponseToMessageId);
// etc...
// }
////////////////////////////////////////////////////////////////////////
#define DEFINE_RESPONSE(classname,\
			imessage,\
			ifunctionname)\
	DefineResponse(imessage,\
				 (pMF)&classname::ifunctionname);
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
#define DEFINE_REFLEX(classname,\
				nvar,\
				ncondittion,\
				nfunctionname)\
	DefineReflex(nvar,\
			  ncondittion,\
			  (pMF)&classname::nfunctionname);
//////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
#define MAKE_RESPONSE(classname,\
			imessage,\
			ifunctionname)\
	MakeResponse(imessage,\
				 (pMF)&classname::ifunctionname);
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
#define MAKE_REFLEX(classname,\
				nvar,\
				ncondittion,\
				nfunctionname)\
	MakeReflex(nvar,\
			  ncondittion,\
			  (pMF)&classname::nfunctionname);
//////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
impulse::impulse(unsigned int Message,pMF pFunc) : NEXUS()
{  // void (impulse::*pFunc)()):NEXUS()
impulseconstructors++;
CONDITION=Message;
RESPOND=pFunc;
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
impulse::impulse(impulse *i):NEXUS()
{
impulseconstructors++;
RESPOND=i->RESPOND;
VAR=i->VAR;
CONDITION=i->CONDITION;
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//impulse::impulse():NEXUS()
//{
//impulseconstructors++;
//VAR=NULL;
//CONDITION=NULL;
//RESPOND=NULL;
//};
///////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
impulse::impulse(unsigned int *Var,
			 unsigned int Condition,
			 pMF pFunc):NEXUS()
{
impulseconstructors++;
VAR=Var;
CONDITION=Condition;
RESPOND=pFunc;
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void impulse::makeresponse(unsigned int Message, pMF pFunc)
{
CONDITION=Message;
RESPOND=pFunc;
VAR=NULL;
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void impulse::respond(NEXUS *pN)
{
 (pN->*RESPOND)();//This works but all RESPOND functions must be NEXUS
 //(this->*RESPOND)();
 //RESPOND();
 //return(RESPOND);
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
impulse::~impulse()
{
impulsedestructors++;
//cout<<"\n"<<"IN impulse Destructor"<<"\n";
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void impulse::makereflex(unsigned int *Var,
			  unsigned int Condition,
			  pMF pFunc)
{
VAR=Var;
CONDITION=Condition;
RESPOND=pFunc;
}
///////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
Reactor::Reactor():ROTI()
{
coordinatorconstructors++;
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
impulse * Reactor::DefineResponse(unsigned int Message,
				 pMF pFunc)
{
 impulse* newimpulse = new impulse(Message,pFunc);
 PUT_MAXIMUS(newimpulse);
 return(newimpulse);
}
///////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////

impulse * Reactor::DefineReflex(unsigned int *Var,
				unsigned int Condition,
				pMF pFunc)
{
 impulse* newnatural = new impulse(Var,Condition,pFunc);
 PUT_MAXIMUS(newnatural);
 return(newnatural);
}
///////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void Reactor::Respond(MSG *Msg,NEXUS *pN)
{
if(PRIMUS!=NULL)
 {
  impulse* current = (impulse *)PRIMUS;
  for(;;)
	  {
		if(current->CONDITION==Msg->message){
		current->respond(pN);
		break;
		}
		if(current->next == NULL)break;
		current=(impulse *)current->next;
	  }
  }
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void Reactor::Reflex(NEXUS *pN)
{
if(PRIMUS)
 {
  impulse* current =(impulse *) PRIMUS;
  for(;;)
	  {
		if(current->CONDITION==*(current->VAR)){
		current->respond(pN);
		}
		if(current->next == NULL)break;
		current=(impulse *)current->next;
	  }
  }
}
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
Reactor::~Reactor()
{
impulse *temp;
coordinatordestructors++;

if((PRIMUS)&&(MAXIMUS))//if not emptty
{
 NEXUS *temp=PRIMUS;

 for(impulse *current=(impulse *)PRIMUS;         //start on PRIMUS
             current;                 //Stop after Maximus
                          )          //increment current
    {
     current=(impulse *)current->next;
     delete temp;
     temp=current;
    }
}
//cout<<"\n"<<"IN Reactor Destructor"<<"\n";
//NEXUS* NEXT;
//NEXUS* current = PRIMUS;//IDLE->PRIMUS;
//if(PRIMUS!=NULL)//IDLE->PRIMUS!=NULL)
 //{
  //for(;;)
	 // {
	 // NEXT=current->next;
	 //	if(current!=NULL)delete current;
	  //	if(current == NULL)break;
	  //	current=NEXT;
	  //	}
  //}

}
///////////////////////////////////////////////////////////////////

Coordinator::Coordinator()
{
ACTIVE=new Reactor();
INACTIVE=new Reactor();
}
///////////////////////////////////////////////////////////////////
impulse * Coordinator::MakeResponse(unsigned int Message,
				 pMF pFunc)
{
 impulse* newimpulse = ACTIVE->DefineResponse(Message,pFunc);
 return(newimpulse);
}
///////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
impulse * Coordinator::MakeReflex(unsigned int *Var,
				unsigned int Condition,
				pMF pFunc)
{
 impulse *newreflex;
 newreflex=ACTIVE->DefineReflex(Var,Condition,pFunc);
 return(newreflex);
}
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
void Coordinator::Activate(impulse *i)
{
//this method of activation is slower than another type that would take
//up more memory so this type takes up less memory
//so this is the small slow version
//1 see if i is inactive
//2 see if i is already active
impulse *current;
//iterate through the inactive list
for(current=(impulse *)INACTIVE->PRIMUS;
	 current;
	 current=(impulse *)current->next)
//see if i is in here
if(i==current)
{
//if so iterate through the active list
for(current=(impulse *)ACTIVE->PRIMUS;
	 current;
	 current=(impulse *)current->next)
	 if(i==current)return;//get out if i is already in the active  list
	 //if we made it here i aint already active
INACTIVE->EXTRACT_NEXUS(i);//if not active get i out of inactive list
ACTIVE->PUT_MAXIMUS(i);//and put i in active list
break;
}
}
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
void Coordinator::Deactivate(impulse *i)
{
//1 see if i is active
//2 see if i is already inactive
impulse *current;
for(current=(impulse *)ACTIVE->PRIMUS;
	 current;
	 current=(impulse *)current->next)
if(i==current)
{
for(current=(impulse *)INACTIVE->PRIMUS;
	 current;
	 current=(impulse *)current->next)if(i==current)return;
ACTIVE->EXTRACT_NEXUS(i);
INACTIVE->PUT_MAXIMUS(i);
break;
}
}


///////////////////////////////////////////////////////////////////////////
void Coordinator::Respond(MSG *Msg,NEXUS *pN)
{
 ACTIVE->Respond(Msg,pN);
};
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
void Coordinator::Reflex(NEXUS *pN)
{
 ACTIVE->Reflex(pN);
}
///////////////////////////////////////////////////////////////////////////


Coordinator::~Coordinator()
{
if(ACTIVE)delete ACTIVE;
if(INACTIVE)delete INACTIVE;
}
